#!/usr/bin/env python
# -*- coding: utf-8 -*-
def predict_final(prefix=None,
                  assignment=None,
                  tutorial=None,
                  midterm=None,
                  takehome=None,
                  final=None):
    """ Predictor for Final from model/59db76eb9b356c2c97004804

        Predictive model by BigML - Machine Learning Made Easy
    """
    if (midterm is None):
        return {"prediction":0.38343}
    if (midterm > 77.08667):
        if (takehome is None):
            return {"prediction":20.38342}
        if (takehome > 106.945):
            return {"prediction":3.43945}
        if (takehome <= 106.945):
            if (tutorial is None):
                return {"prediction":22.41332}
            if (tutorial > 78.665):
                return {"prediction":23.88547}
            if (tutorial <= 78.665):
                return {"prediction":8.56289}
    if (midterm <= 77.08667):
        if (midterm > 48.75):
            if (takehome is None):
                return {"prediction":-4.5295}
            if (takehome > 53.795):
                if (midterm > 73.44):
                    if (takehome > 73.795):
                        return {"prediction":-13.82749}
                    if (takehome <= 73.795):
                        return {"prediction":-3.41771}
                if (midterm <= 73.44):
                    if (assignment is None):
                        return {"prediction":-0.71945}
                    if (assignment > 82.74):
                        if (tutorial is None):
                            return {"prediction":-3.97172}
                        if (tutorial > 103.945):
                            if (tutorial > 104.835):
                                return {"prediction":-0.08074}
                            if (tutorial <= 104.835):
                                return {"prediction":1.25414}
                        if (tutorial <= 103.945):
                            if (midterm > 62.5):
                                if (midterm > 65.31):
                                    return {"prediction":-4.37529}
                                if (midterm <= 65.31):
                                    return {"prediction":4.4972}
                            if (midterm <= 62.5):
                                if (tutorial > 95.71):
                                    return {"prediction":-14.0932}
                                if (tutorial <= 95.71):
                                    return {"prediction":-1.74541}
                    if (assignment <= 82.74):
                        if (tutorial is None):
                            return {"prediction":7.50115}
                        if (tutorial > 96.79):
                            return {"prediction":-0.71028}
                        if (tutorial <= 96.79):
                            return {"prediction":10.35668}
            if (takehome <= 53.795):
                return {"prediction":-13.76724}
        if (midterm <= 48.75):
            if (takehome is None):
                return {"prediction":-15.51536}
            if (takehome > 58.89):
                return {"prediction":-18.93268}
            if (takehome <= 58.89):
                if (tutorial is None):
                    return {"prediction":-5.65621}
                if (tutorial > 77.095):
                    return {"prediction":-7.01337}
                if (tutorial <= 77.095):
                    return {"prediction":-0.79237}


def predict(prefix=None,
            assignment=None,
            tutorial=None,
            midterm=None,
            takehome=None,
            final=None):
    prediction = predict_final(prefix=prefix, assignment=assignment, tutorial=tutorial, midterm=midterm, takehome=takehome, final=final)
    prediction.update({"weight": 0.09621})
    return prediction


# -*- coding: utf-8 -*-
#!/usr/bin/env python
#
# Copyright 2014-2018 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Base class for predictions' REST calls

   https://bigml.com/developers/predictions

"""

try:
    import simplejson as json
except ImportError:
    import json


from bigml.resourcehandler import ResourceHandler
from bigml.resourcehandler import (check_resource_type, get_prediction_id,
                                   check_resource, get_resource_id,
                                   get_resource_type)
from bigml.constants import SUPERVISED_PATHS, TINY_RESOURCE, PREDICTION_PATH


class PredictionHandler(ResourceHandler):
    """This class is used by the BigML class as
       a mixin that provides the REST calls models. It should not
       be instantiated independently.

    """
    def __init__(self):
        """Initializes the PredictionHandler. This class is intended to be
           used as a mixin on ResourceHandler, that inherits its
           attributes and basic method from BigMLConnection, and must not be
           instantiated independently.

        """
        self.prediction_url = self.prediction_url + PREDICTION_PATH

    def create_prediction(self, model, input_data=None,
                          args=None, wait_time=3, retries=10):
        """Creates a new prediction.
           The model parameter can be:
            - a simple tree model
            - a simple logistic regression model
            - an ensemble
            - a deepnet
            - a fusion
           Note that the old `by_name` argument has been deprecated.

        """
        model_id = None

        resource_type = get_resource_type(model)
        if resource_type not in SUPERVISED_PATHS:
            raise Exception("A supervised model resource id is needed"
                            " to create a prediction. %s found." %
                            resource_type)

        model_id = get_resource_id(model)
        if model_id is not None:
            check_resource(model_id,
                           query_string=TINY_RESOURCE,
                           wait_time=wait_time, retries=retries,
                           raise_on_error=True, api=self)

        if input_data is None:
            input_data = {}
        create_args = {}
        if args is not None:
            create_args.update(args)
        create_args.update({
            "input_data": input_data})
        if model_id is not None:
            create_args.update({
                "model": model_id})

        body = json.dumps(create_args)
        return self._create(self.prediction_url, body,
                            verify=self.verify_prediction)

    def get_prediction(self, prediction, query_string=''):
        """Retrieves a prediction.

        """
        check_resource_type(prediction, PREDICTION_PATH,
                            message="A prediction id is needed.")
        prediction_id = get_prediction_id(prediction)
        if prediction_id:
            return self._get("%s%s" % (self.url, prediction_id),
                             query_string=query_string)

    def list_predictions(self, query_string=''):
        """Lists all your predictions.

        """
        return self._list(self.prediction_url, query_string)

    def update_prediction(self, prediction, changes):
        """Updates a prediction.

        """
        check_resource_type(prediction, PREDICTION_PATH,
                            message="A prediction id is needed.")
        prediction_id = get_prediction_id(prediction)
        if prediction_id:
            body = json.dumps(changes)
            return self._update("%s%s" % (self.url, prediction_id), body)

    def delete_prediction(self, prediction):
        """Deletes a prediction.

        """
        check_resource_type(prediction, PREDICTION_PATH,
                            message="A prediction id is needed.")
        prediction_id = get_prediction_id(prediction)
        if prediction_id:
            return self._delete("%s%s" % (self.url, prediction_id))



# -*- coding: utf-8 -*-
#!/usr/bin/env python
#
# Copyright 2014-2018 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Predicates structure for the BigML local AnomalyTree

This module defines an auxiliary Predicates structure that is used in the
AnomalyTree to save the node's predicates info.

"""
from bigml.predicate import Predicate

class Predicates(object):
    """A list of predicates to be evaluated in an anomaly tree's node.

    """
    def __init__(self, predicates_list):
        self.predicates = []
        for predicate in predicates_list:
            if predicate is True:
                self.predicates.append(True)
            else:
                self.predicates.append(
                    Predicate(predicate.get('op'),
                              predicate.get('field'),
                              predicate.get('value'),
                              predicate.get('term')))

    def to_rule(self, fields, label='name'):
        """ Builds rule string from a predicates list

        """
        return " and ".join([predicate.to_rule(fields, label=label) for
                             predicate in self.predicates
                             if not isinstance(predicate, bool)])

    def apply(self, input_data, fields):
        """ Applies the operators defined in each of the predicates to
            the provided input data

        """

        return all([predicate.apply(input_data, fields) for
                    predicate in self.predicates
                    if isinstance(predicate, Predicate)])



# -*- coding: utf-8 -*-
#!/usr/bin/env python
#
# Copyright 2015-2018 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Item object for the Association resource.

   This module defines each item in an Association resource.
"""

from bigml.associationrule import SUPPORTED_LANGUAGES
from bigml.predicate import term_matches, item_matches

class Item(object):
    """ Object encapsulating an Association resource item as described in
    https://bigml.com/developers/associations

    """

    def __init__(self, index, item_info, fields):
        self.index = index
        self.complement = item_info.get('complement', False)
        self.complement_index = item_info.get('complement_index')
        self.count = item_info.get('count')
        self.description = item_info.get('description')
        self.field_id = item_info.get('field_id')
        self.field_info = fields[self.field_id]
        self.name = item_info.get('name')
        self.bin_end = item_info.get('bin_end')
        self.bin_start = item_info.get('bin_start')

    def out_format(self, language="JSON"):
        """Transforming the item structure to a string in the required format

        """
        if language in SUPPORTED_LANGUAGES:
            return getattr(self, "to_%s" % language)()
        return self

    def to_csv(self):
        """Transforming the item to CSV formats

        """
        output = [self.complement, self.complement_index, self.count,
                  self.description, self.field_info['name'], self.name,
                  self.bin_end, self.bin_start]
        return output

    def to_json(self):
        """Transforming the item relevant information to JSON

        """
        item_dict = {}
        item_dict.update(self.__dict__)
        del item_dict["field_info"]
        del item_dict["complement_index"]
        del item_dict["index"]
        return item_dict

    def to_lisp_rule(self):
        """Returns the LISP flatline expression to filter this item

        """
        flatline = ""
        if self.name is None:
            return u"(missing? (f %s))" % self.field_id
        field_type = self.field_info['optype']
        if field_type == "numeric":
            start = self.bin_end if self.complement else \
                self.bin_start
            end = self.bin_start if self.complement else \
                self.bin_end
            if start is not None and end is not None:
                if start < end:
                    flatline = u"(and (< %s (f %s)) (<= (f %s) %s))" % \
                        (start, self.field_id, self.field_id, end)
                else:
                    flatline = u"(or (> (f %s) %s) (<= (f %s) %s))" % \
                        (self.field_id, start, self.field_id, end)
            elif start is not None:
                flatline = u"(> (f %s) %s)" % (self.field_id, start)
            else:
                flatline = u"(<= (f %s) %s)" % (self.field_id, end)
        elif field_type == "categorical":
            operator = u"!=" if self.complement else u"="
            flatline = u"(%s (f %s) %s)" % (
                operator, self.field_id, self.name)
        elif field_type == "text":
            operator = u"=" if self.complement else u">"
            options = self.field_info['term_analysis']
            case_insensitive = not options.get('case_sensitive', False)
            case_insensitive = u'true' if case_insensitive else u'false'
            language = options.get('language')
            language = u"" if language is None else u" %s" % language
            flatline = u"(%s (occurrences (f %s) %s %s%s) 0)" % (
                operator, self.field_id, self.name,
                case_insensitive, language)
        elif field_type == 'items':
            operator = u"!" if self.complement else u""
            flatline = u"(%s (contains-items? %s %s))" % (
                operator, self.field_id, self.name)
        return flatline

    def describe(self):
        """Human-readable description of a item_dict

        """
        description = ""
        if self.name is None:
            return "%s is %smissing" % (
                self.field_info['name'], "not " if self.complement else "")
        field_name = self.field_info['name']
        field_type = self.field_info['optype']
        if field_type == "numeric":
            start = self.bin_end if self.complement else \
                self.bin_start
            end = self.bin_start if self.complement else \
                self.bin_end
            if start is not None and end is not None:
                if start < end:
                    description = "%s < %s <= %s" % (start,
                                                     field_name,
                                                     end)
                else:
                    description = "%s > %s or <= %s" % (field_name,
                                                        start,
                                                        end)
            elif start is not None:
                description = "%s > %s" % (field_name, start)
            else:
                description = "%s <= %s" % (field_name, end)
        elif field_type == "categorical":
            operator = "!=" if self.complement else "="
            description = "%s %s %s" % (field_name, operator, self.name)
        elif field_type in ["text", "items"]:
            operator = "excludes" if self.complement else "includes"
            description = "%s %s %s" % (field_name, operator, self.name)
        else:
            description = self.name
        return description

    def matches(self, value):
        """ Checks whether the value is in a range for numeric fields or
        matches a category for categorical fields.

        """
        field_type = self.field_info['optype']
        if value is None:
            return self.name is None
        if field_type == "numeric" and (
                self.bin_end is not None or self.bin_start is not None):
            if self.bin_start is not None and self.bin_end is not None:
                result = self.bin_start <= value <= self.bin_end
            elif self.bin_end is not None:
                result = value <= self.bin_end
            else:
                result = value >= self.bin_start
        elif field_type == 'categorical':
            result = self.name == value
        elif field_type == 'text':
            # for text fields, the item.name or the related term_forms should
            # be in the considered value
            all_forms = self.field_info['summary'].get('term_forms', {})
            term_forms = all_forms.get(self.name, [])
            terms = [self.name]
            terms.extend(term_forms)
            options = self.field_info['term_analysis']
            result = term_matches(value, terms, options) > 0
        elif field_type == 'items':
            # for item fields, the item.name should be in the considered value
            # surrounded by separators or regexp
            options = self.field_info['item_analysis']
            result = item_matches(value, self.name, options) > 0
        if self.complement:
            result = not result
        return result



import math
import numpy as np

from bigml.laminar.constants import NUMERIC, CATEGORICAL


MODE_CONCENTRATION = 0.1
MODE_STRENGTH = 3

MEAN = "mean"
STANDARD_DEVIATION = "stdev"

ZERO = "zero_value"
ONE = "one_value"

def index(alist, value):
    try:
        return alist.index(value)
    except ValueError:
        return None

def one_hot(vector, possible_values):
    idxs = list(enumerate(index(possible_values, v) for v in vector))
    valid_pairs = filter(lambda x: x[1] is not None, idxs)
    outvec = np.zeros((len(idxs), len(possible_values)), dtype=np.float32)
    for v in valid_pairs:
        outvec[v[0], v[1]] = 1
    return outvec

def standardize(vector, mn, stdev):
    newvec = vector - mn

    if stdev > 0:
        newvec = newvec / stdev

    fill_dft = lambda x: 0.0 if math.isnan(x) else x
    newvec = np.vectorize(fill_dft)(newvec)
    return newvec

def binarize(vector, zero, one):
    if one == 0.0:
        vector[vector == one] = 1.0
        vector[(vector != one) & (vector != 1.0)] = 0.0
    else:
        vector[vector != one] = 0.0
        vector[vector == one] = 1.0

    return vector

def moments(amap):
    return amap[MEAN], amap[STANDARD_DEVIATION]

def bounds(amap):
    return amap[ZERO], amap[ONE]

def transform(vector, spec):
    vtype = spec['type']

    if vtype == NUMERIC:
        if STANDARD_DEVIATION in spec:
            mn, stdev = moments(spec)
            output = standardize(vector, mn, stdev)
        elif ZERO in spec:
            low, high = bounds(spec)
            output = binarize(vector, low, high)
        else:
            raise ValueError("'%s' is not a valid numeric spec!" % str(spec))
    elif vtype == CATEGORICAL:
        output = one_hot(vector, spec['values'])
    else:
        raise ValueError("'%s' is not a valid spec type!" % vtype)
    return output


def tree_predict(tree, point):
    node = tree[:]

    while node[-1] is not None:
        if point[node[0]] <= node[1]:
            node = node[2]
        else:
            node = node[3]

    return node[0]


def get_embedding(X, model):
    if isinstance(model, list):
        preds = None
        for tree in model:
            tree_preds = []
            for row in X:
                tree_preds.append(tree_predict(tree, row))

            if preds is None:
                preds = np.array(tree_preds, dtype='float64')
            else:
                preds += np.array(tree_preds, dtype='float64')

        if len(preds[0]) > 1:
            preds /= preds.sum(axis=1, keepdims=True)
        else:
            preds /= len(model)

        return preds
    else:
        raise ValueError("Model is unknown type!")


def tree_transform(X, trees):
    outdata = None

    for feature_range, model in trees:
        sidx, eidx = feature_range
        inputs = X[:, sidx:eidx]
        outarray = get_embedding(inputs, model)
        if outdata is not None:
            outdata = np.c_[outdata, outarray]
        else:
            outdata = outarray
    return np.c_[outdata, X]


def preprocess(columns, specs):
    outdata = None

    for spec in specs:
        column = columns[spec['index']]

        if spec['type'] == NUMERIC:
            column = np.asarray(column, dtype=np.float32)

        outarray = transform(column, spec)
        if outdata is not None:
            outdata = np.c_[outdata, outarray]
        else:
            outdata = outarray

    return outdata
